@@ -12,7 +12,7 @@ generate insights.
 
 
 __author__ = 'SciBite DataScience'
-__version__ = '0.5.4'
+__version__ = '0.5.6'
 __copyright__ = '(c) 2022, SciBite Ltd'
 __license__ = 'Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License'
 
@@ -20,6 +20,7 @@ __license__ = 'Creative Commons Attribution-NonCommercial-ShareAlike 4.0 Interna
 import os
 import json
 import requests
+import itertools
 import nltk.data
 import urllib.parse
 import termite_toolkit.termite as termite
@@ -133,9 +134,12 @@ class SciBiteAIClient():
 		:param string res_format: The format of the response (by default it's scibite json)
 		:param boolean hits_only: If set to true will return the hits only and not the original
 		sentence(s) By default it is set to false"""
-		req = '/ner/predict_sentences'
-		data = {'models': models, 'sentences': sent, 'res_format': res_format, 'hits_only': hits_only}
-		r = requests.post('http://' + self.scibiteai_url + req + self.encode_data_for_url(data), auth=self.scibiteai_auth)
+		if type(models) == list:
+			models = ','.join(models)
+		hits_only = 'true' if hits_only else 'false'
+		req = f'/ner/predict_sentences?models={models}&res_format={res_format}&hits_only={hits_only}'
+		data = {'sentences': sent}
+		r = requests.post('http://' + self.scibiteai_url + req, json=data, auth=self.scibiteai_auth)
 		return r.json()
 
 	def ner_from_file(self, models, file, res_format='scibite', hits_only=False):
@@ -155,12 +159,11 @@ class SciBiteAIClient():
 			files =  {"file": (file_name, file_obj)},auth=self.scibiteai_auth)
 		
 		return r.json()['task_id']
-	
+
 	###
 	#QA functionality
 	###
 
-
 	def qa_from_json(self, model, file):
 		"""
 		Pass a file containing SQuAD formatted json questions/contexts to answer said questions.
@@ -405,12 +408,60 @@ def markup(docjsonx, normalisation='id', substitute=True, wrap=False,
 	return results
 
 
+def pairwise_markup(docjsonx, pairwise_types_a, pairwise_types_b, normalisation='id', 
+					wrap=False, wrapChars=('{!', '!}'), substitute=True, replacementDict=None):
+	'''
+	Receives TERMite docjsonx, returns a dictionary with pairwise TERMited substitutions.
+
+	:param docjsonx: JSON string generated by TERMite. Must be docjsonx.
+	:param array(str) pairwise_types_a: list of VOCABs to be found on one side of the pairwise relationships
+	:param array(str) pairwise_types_b: list of VOCABS to be found on the other side of the pairwise relationships
+	:param str normalisation: Type of normalisation to substitute/add (must be 'id', 'type', 'name', 'typeplusname' or 'typeplusid')
+	:param bool substitute: Whether to replace the found term (or add normalisation alongside)
+	:param bool wrap: Whether to wrap found hits with 'bookends'
+	:param tuple(str) wrapChars: Tuple of length 2, containing strings to insert at start/end of found hits
+	:param dict replacementDict: Dictionary with <VOCAB>:<string_to_replace_hits_in_vocab>. '~ID~' will be replaced with the entity id,
+	and '~TYPE~' will be replaced with the vocab name. Example: {'GENE':'ENTITY_~TYPE~_~ID~'} would result in BRCA1 -> ENTITY_GENE_BRCA1
+	replacementDict supercedes normalisation. ~NAME~ can also be used to get the preferred name.
+	:return dict: a dictionary containing entity combinations to their respective masked sentences
+	'''
+
+	output = {}
+	ent_id_to_hit_json = {}
+	pairwise_ids_a = []
+	pairwise_ids_b = []
+	try:
+		for hit in docjsonx[0]['termiteTags']:
+			if hit['entityType'] in pairwise_types_a:
+				pairwise_ids_a.append(hit['hitID'])
+			elif hit['entityType'] in pairwise_types_b:
+				pairwise_ids_b.append(hit['hitID'])
+			else:
+				continue
+			ent_id_to_hit_json[hit['hitID']] = hit
+	except TypeError:
+		raise('Error retrieving results from TERMite')
+	except KeyError:
+		pass
+
+	combos = itertools.product(pairwise_ids_a, pairwise_ids_b)
+
+	for combo in combos:
+		termiteTags = [ent_id_to_hit_json[combo[0]], ent_id_to_hit_json[combo[1]]]
+		docjsonx[0]['termiteTags'] = termiteTags
+		output[combo] =markup(docjsonx, vocabs=pairwise_types_a+pairwise_types_b, normalisation=normalisation, 
+					wrap=wrap, wrapChars=wrapChars, substitute=substitute, 
+					replacementDict=replacementDict)[0]['termited_text']
+	return output
+
+
 def text_markup(text, termiteAddr='http://localhost:9090/termite', vocabs=['GENE', 'INDICATION', 'DRUG'],
 				normalisation='id', wrap=False, wrapChars=('{!', '!}'), substitute=True, replacementDict=None,
 				termite_http_user=None, termite_http_pass=None, include_json=False):
 	'''
 	Receives plain text, returns text with TERMited substitutions.
-
+	
+	:param str text: Text in which to markup entities
 	:param str normalisation: Type of normalisation to substitute/add (must be 'id', 'type', 'name', 'typeplusname' or 'typeplusid')
 	:param bool substitute: Whether to replace the found term (or add normalisation alongside)
 	:param bool wrap: Whether to wrap found hits with 'bookends'
@@ -444,6 +495,45 @@ def text_markup(text, termiteAddr='http://localhost:9090/termite', vocabs=['GENE
 					  replacementDict=replacementDict)[0]['termited_text']
 
 
+def pairwise_text_markup(text, pairwise_types_a, pairwise_types_b, termiteAddr='http://localhost:9090/termite',
+				normalisation='id', wrap=False, wrapChars=('{!', '!}'), substitute=True, replacementDict=None,
+				termite_http_user=None, termite_http_pass=None, include_json=False):
+	'''
+	Receives plain text, returns a dictionary with pairwise TERMited substitutions.
+
+	:param str text: Text in which to markup entities
+	:param array(str) pairwise_types_a: list of VOCABs to be found on one side of the pairwise relationships
+	:param array(str) pairwise_types_b: list of VOCABS to be found on the other side of the pairwise relationships
+	:param str normalisation: Type of normalisation to substitute/add (must be 'id', 'type', 'name', 'typeplusname' or 'typeplusid')
+	:param bool substitute: Whether to replace the found term (or add normalisation alongside)
+	:param bool wrap: Whether to wrap found hits with 'bookends'
+	:param tuple(str) wrapChars: Tuple of length 2, containing strings to insert at start/end of found hits
+	:param dict replacementDict: Dictionary with <VOCAB>:<string_to_replace_hits_in_vocab>. '~ID~' will be replaced with the entity id,
+	and '~TYPE~' will be replaced with the vocab name. Example: {'GENE':'ENTITY_~TYPE~_~ID~'} would result in BRCA1 -> ENTITY_GENE_BRCA1
+	replacementDict supercedes normalisation. ~NAME~ can also be used to get the preferred name.
+	:return dict: a dictionary containing entity combinations to their respective masked sentences
+	'''
+	t = termite.TermiteRequestBuilder()
+	t.set_url(termiteAddr)
+	t.set_text(text)
+	t.set_entities(','.join(pairwise_types_a+pairwise_types_b))
+	t.set_subsume(True)
+	t.set_input_format("txt")
+	t.set_output_format("doc.jsonx")
+	if termite_http_pass:
+		t.set_basic_auth(termite_http_user, termite_http_pass, verification=False)
+	docjsonx = t.execute()
+
+	if include_json:
+		return pairwise_markup(docjsonx, pairwise_types_a=pairwise_types_a, pairwise_types_b=pairwise_types_b, 
+								normalisation=normalisation, wrap=wrap, wrapChars=wrapChars, substitute=substitute, 
+								replacementDict=replacementDict), docjsonx
+	else:
+		return pairwise_markup(docjsonx, pairwise_types_a=pairwise_types_a, pairwise_types_b=pairwise_types_b, 
+								normalisation=normalisation, wrap=wrap, wrapChars=wrapChars, substitute=substitute, 
+								replacementDict=replacementDict)
+
+
 def label(docjsonx, vocabs, labelLevel='word'):
 	'''
 	Receives TERMite output docjsonx and returns split text with labels as to what entities are found in that part of the text.
@@ -494,4 +584,4 @@ def label(docjsonx, vocabs, labelLevel='word'):
 
 		results[docIdx] = {'split_text': splitText, 'labels': labels}
 
-	return results
+	return results
\ No newline at end of file
